#include <stdlib.h>
#include <signal.h>

/* GTK includes */
#include <gtk/gtkicontheme.h>
#include <gtk/gtktogglebutton.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkmenuitem.h>
#include <gtk/gtkseparatormenuitem.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkdialog.h>

#include <glib.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if HILDONDESKTOP == 1
# include <libhildondesktop/libhildondesktop.h>
#else
# include <hildon-status-bar-lib/hildon-status-bar-item.h>
#endif

#if HILDON == 1
# include <hildon/hildon-banner.h>
# include <hildon/hildon-sound.h>
#else
# include <hildon-widgets/hildon-banner.h>
# include <hildon-widgets/hildon-system-sound.h>
#endif

#include <stdio.h>
#include <string.h>
#include "osso-statusbar-combo.h"
#include "common.h"
#include "dialog.h"

#if HILDONDESKTOP == 1
HD_DEFINE_PLUGIN(OssoStatusbarCpu,
		osso_statusbar_cpu,
		STATUSBAR_TYPE_ITEM);
#define OSSO_STATUSBAR_CPU_GET_PRIVATE(x) \
	(G_TYPE_INSTANCE_GET_PRIVATE((x), \
				     OSSO_TYPE_STATUSBAR_CPU, \
				     OssoStatusbarCpuPrivate))
#endif

struct ShowRequest {
	GtkWidget *menu;
	GtkWidget *menu_item;
	GSList *list;
	GConfClient *gcc;
	guint id;
};

static gboolean my_timeout(gpointer data);
static void cpuon_gconf_handler(GConfClient *gcc,
				guint con,
				GConfEntry *entry, gpointer user_data);
static void memon_gconf_handler(GConfClient *gcc,
				guint con,
				GConfEntry *entry, gpointer user_data);
static void cpumaxon_gconf_handler(GConfClient *gcc,
				guint con,
				GConfEntry *entry, gpointer user_data);
static void my_gconf_handler(GConfClient * gcc,
			     guint conn,
			     GConfEntry * entry, gpointer user_data);
static void my_gconf_clk_handler(GConfClient * gcc,
			         guint conn,
			         GConfEntry * entry, gpointer user_data);
#ifdef OSSODISPLAY
static void osso_statusbar_cpu_inact(osso_display_state_t state,
				     gpointer user_data);
#else
static void combo_inact(osso_hw_state_t * state, gpointer user_data);
#endif
static void combo_popup(GtkToggleButton *button, gpointer user_data);
static void my_notify(gpointer data, GObject *has_been);

#if HILDONDESKTOP == 1
void osso_statusbar_cpu_finalize(GObject *object);
static void osso_statusbar_cpu_class_init(OssoStatusbarCpuClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);
	object_class->finalize = osso_statusbar_cpu_finalize;
	g_type_class_add_private(klass, sizeof(OssoStatusbarCpuPrivate));
}
#else
void combo_entry(HildonStatusBarPluginFn_st *fn)
{
	g_return_if_fail(fn);
	fn->initialize = combo_initialize;
	fn->destroy = combo_destroy;
	fn->update = combo_update;
	fn->get_priority = combo_get_priority;
}
#endif

#if HILDONDESKTOP == 1
static void osso_statusbar_cpu_init(OssoStatusbarCpu *cpu)
#else
void *combo_initialize(HildonStatusBarItem *item, GtkWidget **button)
#endif
{
	osso_hw_state_t osso_states = {
		.shutdown_ind = FALSE,
		.save_unsaved_data_ind = FALSE,
		.memory_low_ind = FALSE,
		.system_inactivity_ind = TRUE,
		.sig_device_mode_ind = 0
       	};
	GdkPixbuf *icon_pixbuf = NULL;
	GError *error = NULL;
	OssoStatusbarCpuPrivate *info = NULL;
	GtkIconTheme *theme = NULL;

#if HILDONDESKTOP != 1
	g_return_val_if_fail(item, NULL);
#endif

	theme = gtk_icon_theme_get_default();
	icon_pixbuf = gtk_icon_theme_load_icon(theme, HILDON_STATUS_BAR_COMBO_ICON_NAME,
			40, GTK_ICON_LOOKUP_NO_SVG, NULL);
#if HILDONDESKTOP != 1
	if (icon_pixbuf == NULL) {
		gtk_widget_destroy(GTK_WIDGET(item));
		return NULL;
	}
#endif

	/* Take a modifyable copy. */
	if (icon_pixbuf) {
		GdkPixbuf *temp = gdk_pixbuf_copy(icon_pixbuf);
		g_object_unref(icon_pixbuf);
		icon_pixbuf = temp;
	}

#if HILDONDESKTOP == 1
	info = OSSO_STATUSBAR_CPU_GET_PRIVATE(cpu);
#else
	info = g_new0(OssoStatusbarCpuPrivate, 1);
	g_return_val_if_fail(info, NULL);
#endif

	info->osso =
		osso_initialize("osso_statusbar_cpu", VERSION, FALSE,
				NULL);
#if OSSODISPLAY
	osso_hw_set_display_event_cb(info->osso,
				     osso_statusbar_cpu_inact,
				     (gpointer)info);
#else
	osso_hw_set_event_cb(info->osso, &osso_states, combo_inact,
			     (gpointer)info);
#endif

	info->gcc = gconf_client_get_default();

	info->button = gtk_toggle_button_new();

	if (error) {
		g_error_free(error);
		error = NULL;
		info->icon = NULL;
	}
	info->icon = gtk_image_new_from_pixbuf(icon_pixbuf);

	info->show_cpu = gconf_client_get_bool(info->gcc,
			GCONF_PATH_CPU_ON, NULL);
	info->show_mem = gconf_client_get_bool(info->gcc,
			GCONF_PATH_MEM_ON, NULL);
	info->notify_cpumax = gconf_client_get_bool(info->gcc,
			GCONF_PATH_CPUMAX_ON, NULL);

	info->color_cpu = gc_get_int_fall(info->gcc,
				      GCONF_PATH_CPU_COLOR, CPU_DEFAULT);
	info->color_mem = gc_get_int_fall(info->gcc,
				      GCONF_PATH_MEM_COLOR, MEM_DEFAULT);
	info->clock = gconf_client_get_bool(info->gcc,
				            GCONF_PATH_CLOCK_ON, NULL);
	info->maxed_cpu_secs = gc_get_int_fall(info->gcc,
				      GCONF_PATH_MAXED_CPU_SECS, GCONF_PATH_MAXED_CPU_SECS_DEFAULT);

	if (info->maxed_cpu_secs < 5) {
		info->maxed_cpu_secs = GCONF_PATH_MAXED_CPU_SECS_DEFAULT;
	}

	/* icon_pixbuf is owned by info->icon */
	g_object_unref(icon_pixbuf);
	
	gtk_container_add(GTK_CONTAINER(info->button),
			  GTK_WIDGET(info->icon));

#if HILDONDESKTOP == 1
	gtk_container_add(GTK_CONTAINER(cpu),
			  GTK_WIDGET(info->button));
#else
	*button = info->button;
#endif

	memset(info->cpu_samples, 0, sizeof(info->cpu_samples));
	info->maxed_cpu_last_notification = 0;
	info->notify_id = 0;

	if (info->show_cpu || info->show_mem || info->notify_cpumax) {
		info->notify_id = g_timeout_add(2000, my_timeout, info);
	}

	gconf_client_add_dir(info->gcc,
			     GCONF_PATH, GCONF_CLIENT_PRELOAD_NONE, NULL);

	info->gc_notify_id_scpu = gconf_client_notify_add(info->gcc,
			GCONF_PATH_CPU_ON,
			cpuon_gconf_handler,
			info, NULL, NULL);

	info->gc_notify_id_smem = gconf_client_notify_add(info->gcc,
			GCONF_PATH_MEM_ON,
			memon_gconf_handler,
			info, NULL, NULL);
	info->gc_notify_id_cpumax = gconf_client_notify_add(info->gcc,
			GCONF_PATH_CPUMAX_ON,
			cpumaxon_gconf_handler,
			info, NULL, NULL);

	info->gc_notify_id_cpu = gconf_client_notify_add(info->gcc,
						     GCONF_PATH_CPU_COLOR,
						     my_gconf_handler,
						     info, NULL, NULL);
	info->gc_notify_id_mem = gconf_client_notify_add(info->gcc,
						     GCONF_PATH_MEM_COLOR,
						     my_gconf_handler,
						     info, NULL, NULL);
	info->gc_notify_id_clk = gconf_client_notify_add(info->gcc,
						     GCONF_PATH_CLOCK_ON,
						     my_gconf_clk_handler,
						     info, NULL, NULL);
	info->gc_notify_id_maxed_cpu_secs = gconf_client_notify_add(info->gcc,
						     GCONF_PATH_MAXED_CPU_SECS,
						     my_gconf_handler,
						     info, NULL, NULL);

	g_signal_connect(info->button, "toggled", G_CALLBACK(combo_popup),
			 info);

	if (info->clock) {
		info->clock_id = 
			g_signal_connect_after(info->icon, "expose-event",
					G_CALLBACK(clock_overlay),
					NULL);
	}

	gconf_client_clear_cache(info->gcc);
	gtk_widget_show_all(info->button);

#if HILDONDESKTOP == 0
	return info;
#endif
}

#if HILDONDESKTOP == 1
void osso_statusbar_cpu_finalize(GObject *object)
#else
void combo_destroy(void *data)
#endif
{
	osso_hw_state_t osso_states = { FALSE, FALSE, FALSE, FALSE, 0 };
	OssoStatusbarCpuPrivate *info;

#if HILDONDESKTOP == 1
	info = OSSO_STATUSBAR_CPU_GET_PRIVATE(object);
#else
	info = (OssoStatusbarCpuPrivate *)data;
	g_return_if_fail(data);
#endif

	g_source_remove(info->notify_id);
	info->notify_id = 0;

	gconf_client_notify_remove(info->gcc, info->gc_notify_id_scpu);
	gconf_client_notify_remove(info->gcc, info->gc_notify_id_smem);
	gconf_client_notify_remove(info->gcc, info->gc_notify_id_cpu);
	gconf_client_notify_remove(info->gcc, info->gc_notify_id_mem);
	gconf_client_notify_remove(info->gcc, info->gc_notify_id_clk);
	gconf_client_notify_remove(info->gcc, info->gc_notify_id_cpumax);
	gconf_client_notify_remove(info->gcc, info->gc_notify_id_maxed_cpu_secs);
	gconf_client_remove_dir(info->gcc, GCONF_PATH, NULL);

	g_object_unref(G_OBJECT(info->gcc));

	osso_hw_unset_event_cb(info->osso, &osso_states);
	osso_deinitialize(info->osso);

#if HILDONDESKTOP == 0
	g_free(data);
#endif
}

#if HILDONDESKTOP == 0
gint combo_get_priority(void *data)
{
	(void)data;
	return HILDON_STATUS_BAR_CPU_PRIORITY;
}

void combo_update(void *data, gint value1, gint value2, const gchar *str)
{
	(void)value1;
	(void)value2;
	(void)str;
	(void)data;
}
#endif

static gboolean my_timeout(gpointer data)
{
	OssoStatusbarCpuPrivate *info = (OssoStatusbarCpuPrivate *)data;
	GtkImage *image = GTK_IMAGE(info->icon);
	GdkPixbuf *buf = gtk_image_get_pixbuf(image);

	int user, nice, kern, idle, total_change = 0;
	static int old_user, old_nice, old_kern, old_idle;
	int muser, mnice, mkern, midle, mused = 0;
	int w = (info->show_cpu && info->show_mem) ? AREA_W : AREA_W * 2;
	int x = info->show_cpu ? AREA_X + AREA_W / 2 : AREA_X;
	
	if (info->show_cpu || info->notify_cpumax) {
		FILE *foo = fopen("/proc/stat", "r");
		fscanf(foo, "cpu %d %d %d %d", &user, &nice, &kern, &idle);
		fclose(foo);
		total_change = user + nice + kern + idle -
			(old_user + old_nice + old_kern + old_idle);

		if (info->notify_cpumax) {
				guchar user_pct = (user - old_user) * 100 /
				       	total_change;
				guchar cpu_avg = 0;
				gint i;
				gint samples = info->maxed_cpu_secs / 2;

				info->cpu_samples[info->cpu_cur_sample] =
				       	user_pct;

				if (samples >= CPU_NUM_SAMPLES)
				       	samples = CPU_NUM_SAMPLES;
				
				for (i = CPU_NUM_SAMPLES;
					       	i > CPU_NUM_SAMPLES - samples;
					       	i--) {
					cpu_avg += info->cpu_samples[
						(info->cpu_cur_sample + i) %
						CPU_NUM_SAMPLES] / samples;
				}

				info->cpu_cur_sample++;
				if (info->cpu_cur_sample >= CPU_NUM_SAMPLES)
				       	info->cpu_cur_sample = 0;
				
				if (cpu_avg >= MAXED_CPU_NOTIFY_PCT) {
					time_t ti = time(NULL);
					if (info->maxed_cpu_last_notification +
							info->maxed_cpu_secs <= ti) {
						info->maxed_cpu_last_notification = ti;
						hildon_banner_show_information(info->icon, NULL, "Warning: Using excessive CPU");
						hildon_play_system_sound("/usr/share/sounds/ui-wrong_charger.wav");
					}
				}

		}

	}
	if (info->show_mem) {
		FILE *foo = fopen("/proc/meminfo", "r");
		fscanf(foo,
				"MemTotal: %d kB MemFree: %d kB "
				"Buffers: %d kB Cached: %d kB",
				&muser, &mnice, &mkern, &midle);
		fclose(foo);
		mused = muser - mnice - mkern - midle;
	}

	if (info->show_cpu)
		move_left_and_draw(buf,
				(user - old_user) * AREA_H / (total_change),
				(nice - old_nice) * AREA_H / (total_change),
				(kern - old_kern) * AREA_H / (total_change),
				(info->color_cpu >> 16) & 0xff,
				(info->color_cpu >> 8) & 0xff,
				info->color_cpu & 0xff,
				AREA_X, w / 2);
	if (info->show_mem)
		move_left_and_draw(buf,
				mused * AREA_H / muser,
				mkern * AREA_H / muser,
				midle * AREA_H / muser,
				(info->color_mem >> 16) & 0xff,
				(info->color_mem >> 8) & 0xff,
				info->color_mem & 0xff,
				x, (w + 1) / 2);

	gtk_image_set_from_pixbuf(image, buf);

	if (info->show_cpu || info->notify_cpumax) {
		old_user = user;
		old_nice = nice;
		old_kern = kern;
		old_idle = idle;
	}

	return TRUE;
}

static void cpuon_gconf_handler(GConfClient *gcc,
				guint con,
				GConfEntry *entry, gpointer user_data)
{
	OssoStatusbarCpuPrivate *info = user_data;

	(void)gcc;
	(void)con;

	info->show_cpu = gconf_value_get_bool(entry->value);
	
	if (info->show_cpu) {
	       	if (!info->notify_id) {
			info->notify_id =
				g_timeout_add(2000, my_timeout, user_data);
		}
	} else if (!info->show_mem && !info->notify_cpumax && info->notify_id) {
		g_source_remove(info->notify_id);
		info->notify_id = 0;
	}
}

static void memon_gconf_handler(GConfClient *gcc,
				guint con,
				GConfEntry *entry, gpointer user_data)
{
	OssoStatusbarCpuPrivate *info = user_data;

	(void)gcc;
	(void)con;

	info->show_mem = gconf_value_get_bool(entry->value);

	
	if (info->show_mem) {
	       	if (!info->notify_id) {
			info->notify_id =
				g_timeout_add(2000, my_timeout, user_data);
		}
	} else if (!info->show_cpu && !info->notify_cpumax && info->notify_id) {
		g_source_remove(info->notify_id);
		info->notify_id = 0;
	}
}

static void cpumaxon_gconf_handler(GConfClient *gcc,
				guint con,
				GConfEntry *entry, gpointer user_data)
{
	OssoStatusbarCpuPrivate *info = user_data;

	(void)gcc;
	(void)con;

	info->notify_cpumax = gconf_value_get_bool(entry->value);
	
	if (info->notify_cpumax) {
	       	if (!info->notify_id) {
			info->notify_id =
				g_timeout_add(2000, my_timeout, user_data);
		}
	} else if (!info->show_mem && !info->notify_cpumax && info->notify_id) {
		g_source_remove(info->notify_id);
		info->notify_id = 0;
	}
}

static void my_gconf_handler(GConfClient * gcc,
			     guint conn, GConfEntry * entry, gpointer data)
{
	OssoStatusbarCpuPrivate *info = (OssoStatusbarCpuPrivate *)data;

	(void)gcc;
	(void)conn;

	if (strcmp(entry->key, GCONF_PATH_CPU_COLOR) == 0) {
		info->color_cpu = 
			gconf_value_get_int(gconf_entry_get_value(entry));
	} else if (strcmp(entry->key, GCONF_PATH_MEM_COLOR) == 0) {
		info->color_mem = 
			gconf_value_get_int(gconf_entry_get_value(entry));
	} else if (strcmp(entry->key, GCONF_PATH_MAXED_CPU_SECS) == 0) {
		info->maxed_cpu_secs = 
			gconf_value_get_int(gconf_entry_get_value(entry));
	}
}

static void my_gconf_clk_handler(GConfClient * gcc,
			     guint conn, GConfEntry * entry, gpointer data)
{
	OssoStatusbarCpuPrivate *info = (OssoStatusbarCpuPrivate *)data;

	(void)gcc;
	(void)conn;

	info->clock = gconf_value_get_bool(gconf_entry_get_value(entry)
		);
	if (!info->clock && info->clock_id) {
		g_signal_handler_disconnect(info->icon, info->clock_id);
		info->clock_id = 0;
	} else if (info->clock && !info->clock_id) {
		info->clock_id = 
			g_signal_connect_after(info->icon, "expose-event",
				               G_CALLBACK(clock_overlay),
				               NULL);
	}
}

#ifdef OSSODISPLAY
static void osso_statusbar_cpu_inact(osso_display_state_t state,
				     gpointer user_data)
#else
static void combo_inact(osso_hw_state_t * state, gpointer user_data)
#endif
{
	OssoStatusbarCpuPrivate *info = (OssoStatusbarCpuPrivate *)user_data;

#ifdef OSSODISPLAY
	if (state == OSSO_DISPLAY_OFF) {
#else
	if (state->system_inactivity_ind) {
#endif
		if (info->notify_id) {
			g_source_remove(info->notify_id);
			info->notify_id = 0;
		}
	} else {
		if (!info->notify_id) {
			info->notify_id =
				g_timeout_add(2000, my_timeout, user_data);
		}
	}
}

static void combo_popup(GtkToggleButton *button, gpointer user_data)
{
	GtkWidget *menu_item;
	OssoStatusbarCpuPrivate *info = (OssoStatusbarCpuPrivate *)user_data;
	GtkWidget *menu;

	if (!gtk_toggle_button_get_active(button))
		return;

	menu = gtk_menu_new();
	menu_item = gtk_menu_item_new_with_label(_("Show processes"));
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);
	g_signal_connect(menu_item, "activate", G_CALLBACK(top_dialog), NULL);
	
	menu_item = gtk_separator_menu_item_new();
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);

	append_items(GTK_MENU_SHELL(menu), info->gcc, info->osso);

	g_object_weak_ref(G_OBJECT(menu), my_notify, info->button);

	gtk_widget_show_all(menu);
	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, popup_place, info->button, 1,
			gtk_get_current_event_time());

	g_signal_connect(menu,
			"selection-done",
			G_CALLBACK(gtk_object_destroy), NULL);
}
	
static void my_notify(gpointer data, GObject *has_been)
{
	(void)has_been;
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), FALSE);
}
